#include "GLWindow.h"
#include <iostream>
#include <ngl/Camera.h>
#include <ngl/Colour.h>
#include <ngl/Light.h>
#include <ngl/Matrix.h>
#include <ngl/Transformation.h>
#include <ngl/TransformStack.h>
#include <ngl/Material.h>
#include <ngl/NGLInit.h>
#include <ngl/Obj.h>
#include <ngl/Random.h>
#include <ngl/VBOPrimitives.h>
#include <ngl/ShaderManager.h>
#include <boost/foreach.hpp>



//----------------------------------------------------------------------------------------------------------------------
GLWindow::GLWindow(
                   QWidget *_parent
                  ) :
                    QGLWidget(_parent)
{

  // set this widget to have the initial keyboard focus
  setFocus();
  // re-size the widget to that of the parent (in this case the GLFrame passed in on construction)
  this->resize(_parent->size());
  // Now set the initial GLWindow attributes to default values
  // Roate is false
  m_rotate=false;
  // mouse rotation values set to 0
  m_spinXFace=0;
  m_spinYFace=0;
	startTimer(40);
}

// This virtual function is called once before the first call to paintGL() or resizeGL(),
//and then once whenever the widget has been assigned a new QGLContext.
// This function should set up any required OpenGL context rendering flags, defining display lists, etc.

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::initializeGL()
{

  glClearColor(0.4f, 0.4f, 0.4f, 1.0f);			   // Grey Background
  // enable depth testing for drawing
   glEnable(GL_DEPTH_TEST);
   // Now we will create a basic Camera from the graphics library
   // This is a static camera so it only needs to be set once
   // First create Values for the camera position
   ngl::Vector From(20,20,-35);
   ngl::Vector To(0,0,0);
   ngl::Vector Up(0,1,0);
   ngl::NGLInit *Init = ngl::NGLInit::instance();
   #ifdef WIN32
   	glewInit();
   #endif
   Init->initGlew();
   //Init->initVBO();
   m_cam= new ngl::Camera(From,To,Up,ngl::PERSPECTIVE);
   // set the shape using FOV 45 Aspect Ratio based on Width and Height
   // The final two are near and far clipping planes of 0.5 and 10
   m_cam->setShape(45,(float)720.0/576.0,0.5,150,ngl::PERSPECTIVE);
   // now to load the shader and set the values
   // grab an instance of shader manager
   ngl::ShaderManager *shader=ngl::ShaderManager::instance();
   // load a frag and vert shaders
   shader->loadShader("Blinn","shaders/Vertex.vs","shaders/Fragment.fs");
   // set this as the active shader
   shader->useShader("Blinn");
   // now pass the modelView and projection values to the shader
   shader->setShaderParamFromMatrix("Blinn","ViewMatrix",m_cam->getModelView());
   shader->setShaderParamFromMatrix("Blinn","projectionMatrix",m_cam->getProjection());

	 shader->loadShader("Colour","shaders/Colour.vs","shaders/Colour.fs");

	// set this as the active shader
	shader->useShader("Colour");
	shader->setShaderParamFromMatrix("Colour","ViewMatrix",m_cam->getModelView());
	shader->setShaderParamFromMatrix("Colour","projectionMatrix",m_cam->getProjection());

	shader->setShaderParam4f("Colour","Colour",1,1,1,1);


	 // the shader will use the currently active material and light0 so set them
   ngl::Material m(ngl::SILVER);
   m.use();
   ngl::Light *L1 = new ngl::Light(ngl::Vector(0,5,4,1),ngl::Colour(1,1,1,1),ngl::LIGHTLOCAL);
   L1->enable();

	 ngl::VBOPrimitives *prim=ngl::VBOPrimitives::instance();
	 prim->createVBOSphere("sphere",1.0,20);


	 // create the default particle inital position
   ngl::Vector pos(0,0,0);
   ngl::Vector dir;
   ngl::Colour c;
   // loop and create the initial particle list
	 ngl::Random *rng=ngl::Random::instance();

   for(int i=0; i<2; ++i)
   {
     // use the uniform function from the rng class to get a nice
     // even random number
     dir=rng->getRandomVector();
     c=rng->getRandomColour();
     // add the particle to the end of the particle list
     Particle *pt;
     int v=(int)rng->randomPositiveNumber(3);
		 if(v==0)
		 {
        pt= new Sphere(pos,dir,c,M_PI,"Blinn",&m_transformStack);
     }
		 else if( v==1 )
     {
			 pt= new Spark(pos,dir,c,"Colour",&m_transformStack);
     }
		 else
     {
			 pt= new Teapot(pos,dir,c,"Blinn",&m_transformStack);
     }


		 m_particles.push_back(pt);
     m_particles[i]->setMaxLife(rng->randomPositiveNumber(240));
    }
}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget has been resized.
// The new size is passed in width and height.
void GLWindow::resizeGL(
                        int _w,
                        int _h
                       )
{
  glViewport(0,0,_w,_h);
  m_cam->setShape(45,(float)_w/_h,0.5,150,ngl::PERSPECTIVE);
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  (*shader)["Blinn"]->use();
	shader->setShaderParamFromMatrix("Blinn","projectionMatrix",m_cam->getProjection());
  (*shader)["Colour"]->use();
	shader->setShaderParamFromMatrix("Colour","projectionMatrix",m_cam->getProjection());

}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget needs to be painted.
// this is our main drawing routine
void GLWindow::paintGL()
{
  // grab an instance of the shader manager
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  // clear the screen and depth buffer
  shader->useShader("Blinn");

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  // Rotation based on the mouse position for our global
  // transform
  ngl::Transformation trans;
  trans.setRotation(m_spinXFace,m_spinYFace,0);
  // set this in the TX stack
  m_transformStack.setGlobal(trans);

	int size=m_particles.size();
	for (int i=0; i<size; ++i)
	{
		m_particles[i]->draw();
	}
	/*BOOST_FOREACH(Particle *p,m_particles)
	{
		p->draw();
	}*/
}



void GLWindow::addParticles(
														 int _type
									 				 )
{
// create the default particle inital position
ngl::Vector pos=0;
ngl::Vector dir;
ngl::Colour c;
ngl::Random *rng=ngl::Random::instance();
dir=rng->getRandomVector();
c=rng->getRandomColour();
// add the particle to the end of the particle list
Particle *pt;

switch(_type)
{
	case 0 : {pt= new Spark(pos,dir,c,"Colour",&m_transformStack); break;}
	case 1 : {pt= new Sphere(pos,dir,c,M_PI,"Blinn",&m_transformStack); break;}
	case 2 : {pt= new Teapot(pos,dir,c,"Blinn",&m_transformStack); break;}
}

m_particles.push_back(pt);
m_particles[m_particles.size()-1]->setMaxLife(rng->randomPositiveNumber(240));
}


//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseMoveEvent (
                               QMouseEvent * _event
                              )
{
  // note the method buttons() is the button state when event was called
  // this is different from button() which is used to check which button was
  // pressed when the mousePress/Release event is generated
  if(m_rotate && _event->buttons() == Qt::LeftButton)
  {
    m_spinYFace = ( m_spinYFace + (_event->x() - m_origX) ) % 360 ;
    m_spinXFace = ( m_spinXFace + (_event->y() - m_origY) ) % 360 ;
    m_origX = _event->x();
    m_origY = _event->y();
  }
  // re-draw GL
  updateGL();
}


//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mousePressEvent (
                                QMouseEvent * _event
                               )
{
  // this method is called when the mouse button is pressed in this case we
  // store the value where the maouse was clicked (x,y) and set the Rotate flag to true
  if(_event->button() == Qt::LeftButton)
  {
    m_origX = _event->x();
    m_origY = _event->y();
    m_rotate =true;
  }
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseReleaseEvent (
                                  QMouseEvent * _event
                                 )
{
  // this event is called when the mouse button is released
  // we then set Rotate to false
  if (_event->button() == Qt::LeftButton)
  {
    m_rotate=false;
  }
}

void GLWindow::timerEvent(
													 QTimerEvent *_event
												  )
{
	BOOST_FOREACH(Particle *p, m_particles)
	{
	    p->update();
	}
	updateGL();
}

GLWindow::~GLWindow()
{
	std::cout<<"Exit deleting particles\n";
  int size=m_particles.size();
  for(int i=0; i<size; ++i)
  {
     delete m_particles[i];
  }
  m_particles.erase(m_particles.begin(),m_particles.end());
}
